#TODO: support/refactor CITE extraction. This is not yet implemented.

from parsel import Selector
import requests

params = {
    'q': 'blizzard', # search query
    'hl': 'en'       # language of the search   
}


headers = {
    'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36',
    'accept-language': 'en-US,en',
    'referer': f"https://scholar.google.com/scholar?hl={params['hl']}&q={params['q']}"
}


def parsel_get_cite_ids():
    html = requests.get('https://scholar.google.com/scholar', params=params, headers=headers)
    soup = Selector(text=html.text)

    # returns a list of publication ID's -> U8bh6Ca9uwQJ
    return soup.css('.gs_r.gs_or.gs_scl::attr(data-cid)').getall()

def parsel_scrape_cite_results():
    citations = []

    for cite_id in parsel_get_cite_ids():
        html = requests.get(f'https://scholar.google.com/scholar?output=cite&q=info:{cite_id}:scholar.google.com', headers=headers)
        selector = Selector(text=html.text)
         
        # might be issues in the future with extracting data from the table
        if selector.css('#gs_citt').get():
            for result in selector.css('tr'):
                institution = result.xpath('th/text()').get()
                citation = result.xpath('td div/text()').get()

                citations.append({'institution': institution, 'citations': citation})

    return citations



from selenium import webdriver
from selenium_stealth import stealth
from selenium.webdriver.chrome.service import Service
from webdriver_manager.chrome import ChromeDriverManager
from selectolax.lexbor import LexborHTMLParser
from typing import List, Dict, Callable
import pandas as pd
from pathlib import Path

class CustomGoogleScholarTopPublications:
    def __init__(self) -> None:
        pass


    def parse(self, parser: Callable, top_publications_data: Callable):
        '''
        Arugments:
        - parser: Callable. Lexbor parser from google_scholar_top_publication_metrics() function.
        - top_publications_data: Callable. List to append data to. List origin location is google_scholar_top_publication_metrics() function. Line 100.
        
        This function parses data from Google Scholar Organic results and appends data to a List.
        
        It's used by google_scholar_top_publication_metrics().
        
        It returns nothing as it appends data to `top_publications_data`, 
        which appends it to `top_publications_data` List in the google_scholar_top_publication_metrics() function.
        '''

        
        for table in parser.css('tr'):
            try:
                title: str = table.css_first('td.gsc_mvt_t').text()
            except: title = None
            
            try: 
                h5_index: int = table.css_first('a.gs_ibl').text()
            except: h5_index = None
            
            try: 
                h5_index_link: str = f"https://scholar.google.com{table.css_first('a.gs_ibl').attrs['href']}"
            except: h5_index_link = None
            
            try: 
                h5_median: int = table.css_first('span.gs_ibl').text()
            except: h5_median = None
        
            top_publications_data.append({
                'title': title,
                'h5_index': int(h5_index) if h5_index else h5_index,
                'h5_index_link': h5_index_link,
                'h5_median': int(h5_median) if h5_median else h5_median
            })


    def scrape_top_publication_metrics(
            self,
            category: str = '', 
            lang: str = 'en',
            save_to_csv: bool = False, 
            save_to_json: bool = False,
        ) -> List[Dict[str, str]]:
        #TODO add subcategories to subcategory arg
        #TODO: support other languages: lang='spanish' -> 'sp'. https://serpapi.com/google-languages


        '''
        Results comes from: https://scholar.google.com/citations?view_op=top_venues
        
        Returns:
        - title: str
        - h5_index: int
        - h5_index_link: str
        - h5_median: int
        
        Arguments: 
        - save_to_csv: True of False. Default is False. Saves data to CSV file.
        - save_to_json: True of False. Default is False. Saves data to JSON file.
        - lang: str. Language. Defaults to English ('en'). For now, need to be checked yourself. Other languages: https://serpapi.com/google-languages
        - category: str
            Available categories:
            - "bus": Business, Economics & Management
            - "chm": Chemical & Material Sciences
            - "eng": Engineering & Computer Science
            - "med": Health & Medical Sciences
            - "hum": Humanities, Literature & Arts
            - "bio": Life Sciences & Earth Sciences
            - "phy": Physics & Mathematics
            - "soc": Social Sciences
            
        Usage:
        
        from google_scholar_py import CustomGoogleScholarTopPublications
        
        data = google_scholar_top_publication_metrics(category='eng', ...) # sv = swedish
        
        for result in data:
            print(result['title'])
            ...
        '''
        
        # selenium stealth
        options = webdriver.ChromeOptions()
        options.add_argument('--headless')
        options.add_argument('--no-sandbox')
        options.add_argument('--disable-dev-shm-usage')
        
        options.add_experimental_option('excludeSwitches', ['enable-automation'])
        options.add_experimental_option('useAutomationExtension', False)
        
        service = Service(ChromeDriverManager().install())
        driver = webdriver.Chrome(service=service, options=options)
            
        stealth(driver,
            languages=['en-US', 'en'],
            vendor='Google Inc.',
            platform='Win32',
            webgl_vendor='Intel Inc.',
            renderer='Intel Iris OpenGL Engine',
            fix_hairline=True
        )
        
        top_publications_data = []

        if category:
            driver.get(f'https://scholar.google.com/citations?view_op=top_venues&hl={lang}&vq={category}')
            parser = LexborHTMLParser(driver.page_source)
            self.parse(parser=parser, top_publications_data=top_publications_data)
        else: 
            # no vq={category} URL parameter
            driver.get(f'https://scholar.google.com/citations?view_op=top_venues&hl={lang}&vq={category}') # vq='' which will redirect to the page with no applied category
            parser = LexborHTMLParser(driver.page_source)
            self.parse(parser=parser, top_publications_data=top_publications_data)
            
        if save_to_csv:
            pd.DataFrame(data=top_publications_data).to_csv('google_scholar_top_publications_data.csv', 
                                                            index=False, encoding='utf-8')
        if save_to_json:
            pd.DataFrame(data=top_publications_data).to_json('google_scholar_top_publications_data.json', 
                                                            orient='records')
            
        driver.quit()
        return top_publications_data



from selenium import webdriver
from selenium_stealth import stealth
from selenium.webdriver.chrome.service import Service
from webdriver_manager.chrome import ChromeDriverManager
from selectolax.lexbor import LexborHTMLParser
from typing import List, Dict, Callable
import pandas as pd
import re
from pathlib import Path


class CustomGoogleScholarTopMandates:
    def __init__(self) -> None:
        pass
    

    def parse(self, parser: Callable, top_mandates_data: Callable):
        '''
        Arugments:
        - parser: Callable. Lexbor parser from google_scholar_top_mandates_metrics() function.
        - top_mandates_data: Callable. List to append data to. List origin location is google_scholar_top_mandates_metrics() function. Line 100.
        
        This function parses data from Google Scholar Organic results and appends data to a List.
        
        It's used by google_scholar_top_mandates_metrics().
        
        It returns nothing as it appends data to `top_mandates_data`, 
        which appends it to `top_mandates_data` List in the google_scholar_top_mandates_metrics() function.
        '''
        
        for table in parser.css('tr'):
            try:
                # removes "... - cached"
                # https://regex101.com/r/EfljZp/1
                funder: str = re.sub(r'(\s\s-.*)', '', table.css_first('td.gsc_mlt_t').text()) 
            except: funder = None
            
            try:
                link: str = table.css_first('.gsc_mlt_t a').attrs['href']
            except: link = None
            
            try: 
                two_eighteen: int = table.css_first('td:nth-child(4)').text()
                if '-' in two_eighteen:
                    two_eighteen = None
            except: two_eighteen = None
            
            try: 
                twenty_twenty: str = table.css_first('td:nth-child(5)').text()
                if '-' in twenty_twenty:
                    twenty_twenty = None
            except: twenty_twenty = None
            
            try: 
                twenty_one: str = table.css_first('td:nth-child(6)').text()
                if '-' in twenty_one: # missing % in the table
                    twenty_one = None
            except: twenty_one = None
            
            #TODO: fix selector to extract "overall" data
            # `td:nth-child(6)` is not working also
            # try:
            #     overall: str = table.css('.gsc_mlt_n.gsc_mlt_bd').text()
            # except: overall = None
        
            top_mandates_data.append({
                'funder': funder,
                'link': link,
                '2019': two_eighteen,
                '2020': twenty_twenty,
                '2021': twenty_one,
                # 'overall': overall
            })


    def scrape_top_mandates_metrics(
            self,
            save_to_csv: bool = False, 
            save_to_json: bool = False,
            lang: str = 'en'
        ) -> List[Dict[str, str]]:
        #TODO add argument to support other languages https://serpapi.com/google-languages

        '''
        Results comes from: https://scholar.google.com/citations?view_op=mandates_leaderboard
        
        Returns:
        - funder: str
        - link: str
        - 2019: str
        - 2020: str
        - 2021: str
        - overall: str (not extracted at the moment, selector needs to be fixed)
        
        Arguments: 
        - save_to_csv: True of False. Saves data to CSV file. Default is False. 
        - save_to_json: True of False. Saves data to JSON file. Default is False.
        - lang: str. Language. Defaults to English ('en'). For now, need to be checked yourself. Other languages: https://serpapi.com/google-languages
        
        Usage:
        
        from google_scholar_py import CustomGoogleScholarTopMandates
        
        parser = CustomGoogleScholarTopMandates()
        data = parser.scrape_top_mandates_metrics(
            save_to_csv=True,
            save_to_json=False
        )
        print(json.dumps(data, indent=2))

        for result in data:
            print(result['funder'])
            ...
        '''
        
        # selenium stealth
        options = webdriver.ChromeOptions()
        options.add_argument('--headless')
        options.add_argument('--no-sandbox')
        options.add_argument('--disable-dev-shm-usage')
        
        options.add_experimental_option('excludeSwitches', ['enable-automation'])
        options.add_experimental_option('useAutomationExtension', False)
        
        service = Service(ChromeDriverManager().install())
        driver = webdriver.Chrome(service=service, options=options)
            
        stealth(driver,
            languages=['en-US', 'en'],
            vendor='Google Inc.',
            platform='Win32',
            webgl_vendor='Intel Inc.',
            renderer='Intel Iris OpenGL Engine',
            fix_hairline=True
        )
        
        top_mandates_data: list = []

        driver.get(f'https://scholar.google.com/citations?view_op=mandates_leaderboard&hl={lang}')
        parser = LexborHTMLParser(driver.page_source)
        self.parse(parser=parser, top_mandates_data=top_mandates_data)
        
        if save_to_csv:
            pd.DataFrame(data=top_mandates_data).to_csv('google_scholar_top_mandates_data.csv', 
                                                        index=False, encoding='utf-8')
            
        if save_to_json:
            pd.DataFrame(data=top_mandates_data).to_json('google_scholar_top_mandates_data.json', 
                                                        orient='records')
            
        driver.quit()
        return top_mandates_data



from serpapi import GoogleScholarSearch
from urllib.parse import urlsplit, parse_qsl
import itertools

#TODO: support pagination using `async` parameter

class SerpApiGoogleScholarAuthor:
    def __init__(self) -> None:
        pass

    def scrape_google_scholar_author_results(
            self,
            author_id: str,
            api_key: str = None,
            lang: str = 'en',
            parse_articles: bool = False,
            article_pagination: bool = False,
        ):
        
        '''
        Extracts all author data: author info, cited by (table, graph), co-authors, all articles.
        
        Arguments:
        - author_id: author id.
        - api_key: SerpApi api key, https://serpapi.com/manage-api-key
        - lang: language for the search. Default 'en'. More: https://serpapi.com/google-languages
        - parse_articles: parses first page of authour articles. Defalul 'False'.
        - article_pagination: True of False. Enables to parse all articles. Default 'False'.
        
        Usage:
        
        from google_scholar_py.serpapi_backend.author_results import SerpApiGoogleScholarAuthor

        parser = SerpApiGoogleScholarAuthor()
        data = parser.scrape_google_scholar_author_results(
            author_id='nHhtvqkAAAAJ',
            api_key='serpapi_api_key',
            parse_articles=True,
            article_pagination=True,
        )
        
        print(data.keys()) # show available keys

        for article in data['articles']:
            print(article['title'])
        '''
        
        if api_key is None:
            raise Exception('Please enter a SerpApi API key to a `api_key` argument. https://serpapi.com/manage-api-key')
        
        if author_id is None:
            raise Exception('Please enter a author id.')
        
        if api_key and author_id is None:
            raise Exception('Please enter a SerpApi API key to a `api_key`, and a author id to `author_id` arguments.')
        
        params = {
            'api_key': api_key,                  # serpapi api key
            'engine': 'google_scholar_author',   # serpapi parsing engine
            'author_id': author_id,              # search by author id
            'hl': lang                           # language
        }
        
        search = GoogleScholarSearch(params)     # where data extracts on the backend
        
        # parsing ALL articles along with author info
        if parse_articles and article_pagination:
            params['start'] = 0          # page number: 0 is first page, 1 is second, etc.
            params['pagesize'] = 100     # number of articles per page
            
            author_all_articles = []
            
            while True:
                results = search.get_dict()
                
                if 'error' in results:
                    print(results['error'])
                    break
                
                author_all_articles.append(results['articles'])
                
                # check for the `next` page
                if 'next' in results.get('serpapi_pagination', {}):
                    search.params_dict.update(dict(parse_qsl(urlsplit(results['serpapi_pagination']['next']).query)))
                else:
                    break
            
            # remove articles key that creates a nested lists
            results.pop('articles')
            
            # flatten list of all articles
            author_all_articles_flatten = list(itertools.chain(*author_all_articles))
            results['articles'] = author_all_articles_flatten
            
            keys_to_delete = ['search_metadata', 'search_parameters']
            for key_to_delete in keys_to_delete:
                results.pop(key_to_delete)
            
            return results
        
        # parsing ONLY FIRST PAGE of articles along with author info
        if parse_articles:
            search = GoogleScholarSearch(params)
            results = search.get_dict()             # JSON -> Python dict
            
            if 'error' in results:
                raise Exception(results['error'])
            
            keys_to_delete = ['search_metadata', 'search_parameters', 'serpapi_pagination']

            for key_to_delete in keys_to_delete:
                results.pop(key_to_delete)

            return results
        
        # if don't need to parse any articles -> remove them from the JSON
        elif article_pagination or parse_articles is False: 
            search = GoogleScholarSearch(params)
            results = search.get_dict()
            
            if 'error' in results:
                raise Exception(results['error'])
            
            keys_to_delete = ['search_metadata', 'search_parameters', 'articles', 'serpapi_pagination']

            for key_to_delete in keys_to_delete:
                results.pop(key_to_delete)
                
            return results
        



from .organic_results import SerpApiGoogleScholarOrganic
from serpapi import GoogleScholarSearch

#TODO: support extracting actual Cite data, for example Bibtex: shorturl.at/vGNU5

class SerpApiGoogleScholarOrganicCite:
    def __init__(self) -> None:
        pass


    def scrape_google_scholar_cite_results(
            self,
            query: str,
            api_key: str = None,
            lang: str = 'en',
            pagination: bool = False
        ):
        
        '''
        This function extract citations as well as BibTeX, EndNote, RefMan, RefWorks links.
        
        To extract citations, 2 requests has to be made: first for organic results, second for citation data. 
        So if you need to get citations from 1000 articles, 2000 requests would be made accordingly.
        
        Arguments:
        - query: search query
        - api_key: SerpApi api key, https://serpapi.com/manage-api-key
        - lang: language for the search. Default 'en'. More: https://serpapi.com/google-languages
        - pagination: True of False. Enables pagination from all pages. Default 'False'.
        
        Usage:
        
        from google_scholar_py.serpapi_backend.organic_cite_results import SerpApiGoogleScholarOrganicCite
        
        parser = SerpApiGoogleScholarOrganicCite()
        data = parser.scrape_google_scholar_cite_results(
            query='minecraft', 
            api_key='serpapi_api_key', 
            pagination=True
        )
        
        # extracting bottom links
        for result in data:
            for citations in result['links']: 
                print(citations['name']) # or ['link']
        
        # extracting citations
        for result in data:
            for citations in result['citations']: 
                print(citations['title']) # or ['snippet'] 
        '''
        
        if api_key is None:
            raise Exception('Please enter a SerpApi API key to a `api_key` argument. https://serpapi.com/manage-api-key')
        
        #TODO: could be removed as function by itself throw an error if query is missing
        if api_key and query is None:
            raise Exception('Please enter a SerpApi API key to a `api_key`, and a search query to `query` arguments.')
        
        # extract organic results from where citation data will be extracted
        organic_results = SerpApiGoogleScholarOrganic().scrape_google_scholar_organic_results(
            query=query,
            api_key=api_key,
            lang=lang,
            pagination=pagination
        )
        
        cite_results_data = []

        for citation in organic_results:
            params = {
                'api_key': api_key,              # serpapi api key: https://serpapi.com/manage-api-key
                'engine': 'google_scholar_cite', # serpapi parsing engine
                'q': citation['result_id']       # search query
            }
            
            search = GoogleScholarSearch(params) # where data extracts on the backend
            results = search.get_dict()
            
            # removes 2 keys from the JSON response 
            for key_to_delete in ['search_metadata', 'search_parameters']:
                results.pop(key_to_delete)
                
            if 'error' in results:
                raise Exception(results['error'])
                
            cite_results_data.append(results)

        return cite_results_data



from selenium import webdriver
from selenium_stealth import stealth
from selenium.webdriver.chrome.service import Service
from webdriver_manager.chrome import ChromeDriverManager
from selectolax.lexbor import LexborHTMLParser
from parsel import Selector
from typing import List, Dict, Callable
import time, random, re
import pandas as pd
from pathlib import Path

class CustomGoogleScholarProfiles:
    def __init__(self) -> None:
        pass
    

    def parse(self, parser: Callable, profile_results_data: Callable):
        '''
        Arugments:
        - parser: Callable. Lexbor parser from scrape_google_scholar_profiles() function.
        - profile_results_data: Callable. List to append data to. List origin location is scrape_google_scholar_profiles() function. Line 100.
        
        This function parses data from Google Scholar Organic results and appends data to a List.
        
        It's used by scrape_google_scholar_profiles().
        
        It returns nothing as it appends data to `profile_results_data`, 
        which appends it to `profile_results_data` List in the scrape_google_scholar_profiles() function.
        '''
        
        for profile in parser.css('.gs_ai_chpr'):
            try:
                name: str = profile.css_first('.gs_ai_name a').text()
            except: name = None
            
            try:
                link: str = f'https://scholar.google.com{profile.css_first(".gs_ai_name a").attrs["href"]}'
            except: link = None
            
            try:
                affiliations: str = profile.css_first('.gs_ai_aff').text()
            except: affiliations = None
            
            try:
                interests: list = [interest.text() for interest in profile.css('.gs_ai_one_int')]
            except: interests = None
            
            try:
                email: str = profile.css_first('.gs_ai_eml').text()
            except: email = None
            
            try:
                cited_by: int = re.search(r'\d+', profile.css_first('.gs_ai_cby').text()).group() # Cited by 17143 -> 17143
            except: cited_by = None

            profile_results_data.append({
                'name': name,
                'link': link,
                'affiliations': affiliations,
                'interests': interests if interests else None,
                'email': email if email else None,
                'cited_by_count': int(cited_by) if cited_by else None
            })


    def scrape_google_scholar_profiles(
            self, 
            query: str, 
            pagination: bool = False, 
            save_to_csv: bool = False, 
            save_to_json: bool = False
        ) -> List[Dict[str, str]]:
        '''
        Extracts data from Google Scholar Organic Profile resutls page:
        - name: str
        - link: str
        - affiliations: str 
        - email: str
        - cited_by_count: int
        
        Arguments:
        - query: str. Search query. 
        - pagination: bool. Enables or disables pagination. Default is False.
        - save_to_csv: bool. True of False. Default is False.
        - save_to_json: bool. True of False. Default is False.
        
        Usage:
        
        from google_scholar_py import CustomGoogleScholarProfiles
        
        parser = CustomGoogleScholarProfiles()
        data = parser.scrape_google_scholar_profiles(
            query='blizzard',
            pagination=False,
            save_to_csv=True
        )
        print(json.dumps(data, indent=2))
        
        for profile_results in data:
            print(profile_results['name'])
            print(profile_results['email'])
        '''
        
        # selenium stealth
        options = webdriver.ChromeOptions()
        options.add_argument('--headless')
        options.add_argument('--no-sandbox')
        options.add_argument('--disable-dev-shm-usage')
        
        options.add_experimental_option('excludeSwitches', ['enable-automation'])
        options.add_experimental_option('useAutomationExtension', False)
        
        service = Service(ChromeDriverManager().install())
        driver = webdriver.Chrome(service=service, options=options)
        
        stealth(driver,
            languages=['en-US', 'en'],
            vendor='Google Inc.',
            platform='Win32',
            webgl_vendor='Intel Inc.',
            renderer='Intel Iris OpenGL Engine',
            fix_hairline=True
        )
        
        params = {} # stores next page token to add to URL later
        page_num = 0
        profile_results_data = []

        if pagination:
            while True:
                # if next page token appears, add to to URL as URL parameter
                # otherwise, do a search without next page token parameter (Line: 101)
                if params.get('after_author') is None:
                    driver.get(f'https://scholar.google.com/citations?view_op=search_authors&mauthors={query}&hl=en&astart={page_num}')
                    parser = LexborHTMLParser(driver.page_source)
                    
                    #TODO: replace parsel with selectolax completely
                    selector = Selector(text=driver.page_source)  # to check next page token
                
                    self.parse(parser=parser, profile_results_data=profile_results_data)
                    
                    # check if the next arrow button is active by checking 'onclick' attribute
                    if selector.css('.gsc_pgn button.gs_btnPR::attr(onclick)').get():
                        # extracting next page token and passing to 'after_author' query URL parameter
                        params['after_author'] = re.search(r'after_author\\x3d(.*)\\x26', str(selector.css('.gsc_pgn button.gs_btnPR::attr(onclick)').get())).group(1) # -> XB0HAMS9__8J
                        page_num += 10                        # paginate to the next page
                        time.sleep(random.randint(1, 3))      # sleep between paginations
                    else:
                        break
                else:
                    driver.get(f'https://scholar.google.com/citations?view_op=search_authors&mauthors={query}&hl=en&astart={page_num}&after_author={params["after_author"]}')
                    parser = LexborHTMLParser(driver.page_source)
                    
                    #TODO: replace parsel with selectolax completely
                    selector = Selector(text=driver.page_source) # to check next page token
                
                    self.parse(parser=parser, profile_results_data=profile_results_data)
                    
                    if selector.css('.gsc_pgn button.gs_btnPR::attr(onclick)').get():
                        # extracting next page token and passing to 'after_author' query URL parameter
                        params['after_author'] = re.search(r'after_author\\x3d(.*)\\x26', str(selector.css('.gsc_pgn button.gs_btnPR::attr(onclick)').get())).group(1) # -> XB0HAMS9__8J
                        page_num += 10                        # paginate to the next page
                        time.sleep(random.randint(1, 3))      # sleep between paginations
                    else:
                        break
        else:
            # parse single, first page
            driver.get(f'https://scholar.google.com/citations?view_op=search_authors&mauthors={query}&hl=en&astart={page_num}')
            parser = LexborHTMLParser(driver.page_source)
        
            self.parse(parser=parser, profile_results_data=profile_results_data)

        driver.quit()
        
        if save_to_csv:
            pd.DataFrame(data=profile_results_data).to_csv('google_scholar_profile_results_data.csv', 
                                                            index=False, encoding='utf-8')
        if save_to_json:
            pd.DataFrame(data=profile_results_data).to_json('google_scholar_profile_results_data.json', 
                                                            orient='records')
        
        return profile_results_data       



# custom backend
from .custom_backend.organic_search import CustomGoogleScholarOrganic
from .custom_backend.profiles_results import CustomGoogleScholarProfiles
from .custom_backend.author_info_all_articles import CustomGoogleScholarAuthor
from .custom_backend.top_mandates_metrics import CustomGoogleScholarTopMandates
from .custom_backend.top_publications_metrics import CustomGoogleScholarTopPublications

# serpapi backend
from .serpapi_backend.organic_results import SerpApiGoogleScholarOrganic
from .serpapi_backend.profile_results import SerpApiGoogleScholarProfiles
from .serpapi_backend.organic_cite_results import SerpApiGoogleScholarOrganicCite
from .serpapi_backend.author_results import SerpApiGoogleScholarAuthor


from selenium import webdriver
from selenium_stealth import stealth
from selenium.webdriver.chrome.service import Service
from webdriver_manager.chrome import ChromeDriverManager
from selectolax.lexbor import LexborHTMLParser
from typing import List, Dict, Callable
import time, random, re
import pandas as pd
from pathlib import Path


class CustomGoogleScholarOrganic:
    def __init__(self) -> None:
        pass
    

    def parse(self, parser: Callable, organic_results_data: Callable):
        '''
        Arugments:
        - parser:  Lexbor parser from scrape_google_scholar_organic_results() function.
        - organic_results_data: List to append data to. List origin location is scrape_google_scholar_organic_results() function. Line 104.
        
        This function parses data from Google Scholar Organic results and appends data to a List.
        
        It's used by scrape_google_scholar_organic_results().
        
        It returns nothing as it appends data to `organic_results_data`, 
        which appends it to `organic_results_data` List in the scrape_google_scholar_organic_results() function.
        '''
        
        for result in parser.css('.gs_r.gs_or.gs_scl'):
            try:
                title: str = result.css_first('.gs_rt').text()
            except: title = None

            try:
                title_link: str = result.css_first('.gs_rt a').attrs['href']
            except: title_link = None

            try:
                publication_info: str = result.css_first('.gs_a').text()
            except: publication_info = None

            try:
                snippet: str = result.css_first('.gs_rs').text()
            except: snippet = None

            try:
                # if Cited by is present in inline links, it will be extracted
                cited_by_link = ''.join([link.attrs['href'] for link in result.css('.gs_ri .gs_fl a') if 'Cited by' in link.text()])
            except: cited_by_link = None
            
            try:
                # if Cited by is present in inline links, it will be extracted and type cast it to integer
                cited_by_count = int(''.join([re.search(r'\d+', link.text()).group() for link in result.css('.gs_ri .gs_fl a') if 'Cited by' in link.text()]))
            except: cited_by_count = None
            
            try:
                pdf_file: str = result.css_first('.gs_or_ggsm a').attrs['href']
            except: pdf_file = None

            organic_results_data.append({
                'title': title,
                'title_link': title_link,
                'publication_info': publication_info,
                'snippet': snippet if snippet else None,
                'cited_by_link': f'https://scholar.google.com{cited_by_link}' if cited_by_link else None,
                'cited_by_count': cited_by_count if cited_by_count else None,
                'pdf_file': pdf_file
            })

    #TODO: add lang support. https://serpapi.com/google-languages
    def scrape_google_scholar_organic_results(
            self,
            query: str,
            pagination: bool = False,
            save_to_csv: bool = False, 
            save_to_json: bool = False
        ) -> List[Dict[str, str]]:
        '''
        Extracts data from Google Scholar Organic resutls page:
        - title: str
        - title_link: str
        - publication_info: str 
        - snippet: str
        - cited_by_link: str 
        - cited_by_count: int
        - pdf_file: str
        
        Arguments:
        - query: str. Search query. 
        - pagination: bool. Enables or disables pagination. Default is False.
        - save_to_csv: bool. True of False. Default is False.
        - save_to_json: bool. True of False. Default is False.
        
        Usage:
        
        from google_scholar_py import CustomGoogleScholarOrganic

        parser = CustomGoogleScholarOrganic()
        data = parser.scrape_google_scholar_organic_results(
            query='blizzard',
            pagination=False,
            save_to_csv=True
        )
        
        for organic_result in data:
            print(organic_result['title'])
            print(organic_result['pdf_file'])
        '''
        
        # selenium stealth
        options = webdriver.ChromeOptions()
        options.add_argument('--headless')
        options.add_argument('--no-sandbox')
        options.add_argument('--disable-dev-shm-usage')
        
        options.add_experimental_option('excludeSwitches', ['enable-automation'])
        options.add_experimental_option('useAutomationExtension', False) 
        
        service = Service(ChromeDriverManager().install())
        driver = webdriver.Chrome(service=service, options=options)
        
        stealth(driver,
            languages=['en-US', 'en'],
            vendor='Google Inc.',
            platform='Win32',
            webgl_vendor='Intel Inc.',
            renderer='Intel Iris OpenGL Engine',
            fix_hairline=True,
        )
        
        page_num = 0
        organic_results_data = []
    
        # parse all pages
        if pagination:
            while True:
                # parse all pages
                driver.get(f'https://scholar.google.com/scholar?q={query}&hl=en&gl=us&start={page_num}')
                parser = LexborHTMLParser(driver.page_source)
                
                self.parse(parser=parser, organic_results_data=organic_results_data)
                
                # pagination
                if parser.css_first('.gs_ico_nav_next'):  # checks for the "Next" page button
                    page_num += 10                        # paginate to the next page
                    time.sleep(random.randint(1, 3))      # sleep between paginations
                else:
                    break
        else:
            # parse first page only
            driver.get(f'https://scholar.google.com/scholar?q={query}&hl=en&gl=us&start={page_num}')
            parser = LexborHTMLParser(driver.page_source)
        
            self.parse(parser=parser, organic_results_data=organic_results_data)
            
        if save_to_csv:
            pd.DataFrame(data=organic_results_data).to_csv('google_scholar_organic_results_data.csv', 
                                                            index=False, encoding='utf-8')
        if save_to_json:
            pd.DataFrame(data=organic_results_data).to_json('google_scholar_organic_results_data.json', 
                                                            orient='records')
        driver.quit()
        
        return organic_results_data



from parsel import Selector
import requests, json

#TODO: add cited by graph extraction to author script

def parsel_scrape_author_cited_by_graph():
    params = {
        'user': '_xwYD2sAAAAJ',       # user-id
        'hl': 'en'                    # language
    }

    headers = {
        'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36'
    }

    data = {
        'cited_by': [],
        'graph': []
    }

    html = requests.get('https://scholar.google.com/citations', params=params, headers=headers, timeout=30)
    selector = Selector(text=html.text)

    since_year = selector.css('.gsc_rsb_sth~ .gsc_rsb_sth+ .gsc_rsb_sth::text').get().lower().replace(' ', '_')

    for cited_by_public_access in selector.css('.gsc_rsb'):
        data['cited_by'].append({
            'citations_all': cited_by_public_access.css('tr:nth-child(1) .gsc_rsb_sc1+ .gsc_rsb_std::text').get(),
            f'citations_since_{since_year}': cited_by_public_access.css('tr:nth-child(1) .gsc_rsb_std+ .gsc_rsb_std::text').get(),
            'h_index_all': cited_by_public_access.css('tr:nth-child(2) .gsc_rsb_sc1+ .gsc_rsb_std::text').get(),
            f'h_index_since_{since_year}': cited_by_public_access.css('tr:nth-child(2) .gsc_rsb_std+ .gsc_rsb_std::text').get(),
            'i10_index_all': cited_by_public_access.css('tr~ tr+ tr .gsc_rsb_sc1+ .gsc_rsb_std::text').get(),
            f'i10_index_since_{since_year}': cited_by_public_access.css('tr~ tr+ tr .gsc_rsb_std+ .gsc_rsb_std::text').get(),
            'articles': {
                    'available': int(cited_by_public_access.css('.gsc_rsb_m_a:nth-child(1) span::text').get().split(' ')[0]), # to get only digit value
                    'not_available': int(cited_by_public_access.css('.gsc_rsb_m_na div::text').get().split(' ')[0]),          # to get only digit value
                },
            'articles_link': f"https://scholar.google.com{cited_by_public_access.css('#gsc_lwp_mndt_lnk::attr(href)').get()}"
        })
    
    for graph_year, graph_yaer_value in zip(selector.css('.gsc_g_t::text'), selector.css('.gsc_g_al::text')):
        data['graph'].append({
            'year': graph_year.get(),
            'value': int(graph_yaer_value.get())
        })

if __name__ == '__main__':
    print(json.dumps(parsel_scrape_author_cited_by_graph(), indent=2, ensure_ascii=False))



from selenium import webdriver
from selenium_stealth import stealth
from selenium.webdriver.chrome.service import Service
from webdriver_manager.chrome import ChromeDriverManager
from selectolax.lexbor import LexborHTMLParser
from typing import List, Union, Dict
from pathlib import Path


class CustomGoogleScholarAuthor:
    def __init__(self) -> None:
        pass


    def scrape_google_scholar_author_data(
            self,
            user_id: str, 
            parse_articles: bool = False,
            article_pagination: bool = False 
        ) -> Dict[str, List[Union[str, int, None]]]:
        '''
        Extracts data from Google Scholar Author profile page:
        - Info about the author itself
        - Co-authors: name, link, affiliation
        - Author: title, link, authors, publication, cited by, year.
        - Articles: first 100 if pagination is False, or all if pagination is True. 

        Arguments:
        - user_id: str. User ID from Google Scholar profile located in the URL.
        - parse_articles: True of False. If True, extracts first 100 articles. Default False.
        - article_pagination: True of False. If True, extracts beyond first 100 articles. 

        Usage:
        
        from google_scholar_py import CustomGoogleScholarAuthor
        
        parser = CustomGoogleScholarAuthor()
        data = parser.scrape_google_scholar_author_data(
            user_id='nHhtvqkAAAAJ',
            parse_articles=True,
            article_pagination=True
        )
        print(json.dumps(data, indent=2))
        
        print(data['info']) # author info
        print(data['co-authors'])

        for article in data['articles']:
            print(article['title'])
            print(article['cited_by_count'])
            ...
        '''  

        # selenium stealth
        options = webdriver.ChromeOptions()
        options.add_argument('--headless')
        options.add_argument('--no-sandbox')
        options.add_argument('--disable-dev-shm-usage')
        
        options.add_experimental_option('excludeSwitches', ['enable-automation'])
        options.add_experimental_option('useAutomationExtension', False)
        
        service = Service(ChromeDriverManager().install())
        driver = webdriver.Chrome(service=service, options=options)
        
        stealth(driver,
            languages=['en-US', 'en'],
            vendor='Google Inc.',
            platform='Win32',
            webgl_vendor='Intel Inc.',
            renderer='Intel Iris OpenGL Engine',
            fix_hairline=True,
        )
        
        driver.get(f'https://scholar.google.com/citations?user={user_id}&hl=en&gl=us&pagesize=100')
        parser = LexborHTMLParser(driver.page_source)
        
        profile_info = {
            'info': {},
            'co-authors': [], 
            'articles': [][:-1] # [:-1] to not to return the last None element. Weird approach, I know. Revisit in the future.
        }
        
        profile_info['info']['name'] = parser.css_first('#gsc_prf_in').text()
        profile_info['info']['affiliations'] = parser.css_first('.gsc_prf_ila').text()
        profile_info['info']['email'] = parser.css_first('#gsc_prf_ivh').text()
        profile_info['info']['interests'] = [interest.text() for interest in parser.css('#gsc_prf_int .gs_ibl')]
        
        for co_author in parser.css('.gsc_rsb_aa'):
            profile_info['co-authors'].append({
                'name': co_author.css_first('.gsc_rsb_a_desc a').text(),
                'profile_link': f"https://scholar.google.com{co_author.css_first('.gsc_rsb_a_desc a').attrs['href']}",
                'affiliation': co_author.css_first('.gsc_rsb_a_ext').text(),
            })
            
        # extracts only first 100 articles, WITHOUT paginaiton
        if parse_articles:
            # TODO: make a separate function to extract articles
            for index, article in enumerate(parser.css('.gsc_a_tr'), start=1):
                try:
                    article_title = article.css_first('.gsc_a_at').text()
                except: article_title = None
                
                try: 
                    article_link = f"https://scholar.google.com{article.css_first('.gsc_a_at').attrs['href']}"
                except: article_link = None
                
                try:
                    if ',' in article.css_first('.gsc_a_at+ .gs_gray').text():
                        article_authors: List[str] = article.css_first('.gsc_a_at+ .gs_gray').text().split(', ') # list of authors
                    else: article_authors = article.css_first('.gsc_a_at+ .gs_gray').text()           # single authour
                except: article_authors = None
                
                try:
                    article_publication = article.css_first('.gs_gray+ .gs_gray').text()
                except: article_publication = None

                try:
                    cited_by_count = article.css_first('.gsc_a_ac').text() 
                except: cited_by_count = None
                
                try: 
                    publication_year = article.css_first('.gsc_a_hc').text()
                except: publication_year = None

                profile_info['articles'].append({
                    'title': article_title,
                    'link': article_link,
                    'authors': article_authors,
                    'publication': article_publication if article_publication else None,
                    'publication_year': int(publication_year) if publication_year else publication_year or None, # int value or None or empty str
                    'cited_by_count': int(cited_by_count) if cited_by_count else cited_by_count or None # int value or None or empty str
                }) 
        elif parse_articles is False:
            profile_info.pop('articles')

        page_num = 0

        # extracts all articles
        if parse_articles and article_pagination:
            while True:
                driver.get(f'https://scholar.google.com/citations?user={user_id}&hl=en&gl=us&cstart={page_num}&pagesize=100')
                parser = LexborHTMLParser(driver.page_source)
                
                for article in parser.css('.gsc_a_tr'):
                    try:
                        article_title = article.css_first('.gsc_a_at').text()
                    except: article_title = None
                    
                    try: 
                        article_link = f"https://scholar.google.com{article.css_first('.gsc_a_at').attrs['href']}"
                    except: article_link = None
                    
                    try:
                        if ',' in article.css_first('.gsc_a_at+ .gs_gray').text():
                            article_authors: List[str] = article.css_first('.gsc_a_at+ .gs_gray').text().split(', ') # list of authors
                        else: article_authors = article.css_first('.gsc_a_at+ .gs_gray').text()           # single authour
                    except: article_authors = None
                    
                    try:
                        article_publication = article.css_first('.gs_gray+ .gs_gray').text()
                    except: article_publication = None

                    try:
                        cited_by_count = article.css_first('.gsc_a_ac').text() 
                    except: cited_by_count = None
                    
                    try: 
                        publication_year = article.css_first('.gsc_a_hc').text()
                    except: publication_year = None

                    profile_info['articles'].append({
                        'title': article_title,
                        'link': article_link,
                        'authors': article_authors,
                        'publication': article_publication if article_publication else None,
                        'publication_year': int(publication_year) if publication_year else publication_year or None, # int value or None or empty str
                        'cited_by_count': int(cited_by_count) if cited_by_count else cited_by_count or None # int value or None or empty str
                    })

                if parser.css_first('.gsc_a_e'):
                    break
                else:
                    page_num += 100  # paginate to the next page
                    
        # remove articles key if user don't want to extract it
        elif article_pagination and parse_articles is False: 
            profile_info.pop('articles')
            
        return profile_info



from serpapi import GoogleScholarSearch
from urllib.parse import parse_qsl, urlsplit
import itertools


#TODO: support pagination using `async` parameter

class SerpApiGoogleScholarProfiles:
    def __init__(self) -> None:
        pass

    def scrape_google_scholar_profile_results(
            self,
            query: str,
            api_key: str = None,
            lang: str = 'en',
            pagination: bool = False,
        ):
        
        '''
        This function extracts profile results. With or without pagination.
        
        Arguments:
        - query: search query
        - api_key: SerpApi api key, https://serpapi.com/manage-api-key
        - lang: language for the search. Default 'en'. More: https://serpapi.com/google-languages
        - pagination: True of False. Enables pagination from all pages. Default 'False'.
        
        Usage:
        
        from google_scholar_py.serpapi_backend.profile_results import SerpApiGoogleScholarProfiles

        parser = SerpApiGoogleScholarProfiles()
        data = parser.scrape_google_scholar_profile_results(
            query='minecraft', 
            api_key='serpapi_api_key', 
            pagination=True,
        )
        
        print(data[0].keys()) # show available keys
        
        for result in data:
            print(result['title'])
            # get other data
        '''
        
        if api_key is None:
            raise Exception('Please enter a SerpApi API key to a `api_key` argument. https://serpapi.com/manage-api-key')
        
        if api_key and query is None:
            raise Exception('Please enter a SerpApi API key to a `api_key`, and a search query to `query` arguments.')
        
        params = {
            'api_key': api_key,                       # serpapi api key: https://serpapi.com/manage-api-key
            'engine': 'google_scholar_profiles',      # serpapi parsing engine
            'mauthors': query,                        # search query
            'hl': lang                                # language
        }
        
        search = GoogleScholarSearch(params) # where data extracts on the backend
        
        if pagination:
            profile_results_data = []
            
            while True:
                results = search.get_dict()  # JSON -> Python dict
                
                if 'error' in results:
                    print(results['error'])
                    break
                
                profile_results_data.append(results['profiles'])
                
                # check for 'next' page
                if 'next' in results.get('pagination', {}):
                    search.params_dict.update(dict(parse_qsl(urlsplit(results['pagination']['next']).query)))
                else:
                    break
                
            # flatten list
            return list(itertools.chain(*profile_results_data))
        else:
            search = GoogleScholarSearch(params)
            results = search.get_dict()
            
            if 'error' in results:
                raise Exception(results['error'])

            return results['profiles']



from setuptools import setup

README = ''
with open('README.md', 'r', encoding='utf-8') as readme_file:
    README = readme_file.read()

setup(
    name='scrape-google-scholar-py',
    description = 'Extract data from all Google Scholar pages in Python. Sponsored by SerpApi.',
    url='https://github.com/dimitryzub/scrape-google-scholar',
    version='0.3.1',
    license='MIT',
    author='Dmitiry Zub',
    author_email='dimitryzub@gmail.com',
    maintainer='Dmitiry Zub',
    maintainer_email='dimitryzub@gmail.com',
    long_description_content_type='text/markdown',
    long_description=README,
    include_package_data=True,
    python_requires='>=3.10',
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',
        'Topic :: Internet',
        'Natural Language :: English',
        'Topic :: Utilities',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    keywords=[
            'google scholar',
            'serpapi',
            'scraper',
            'python',
            'python google scholar',
            'python google scholar api',
            'web scraping',
            'python web scraping',
            'research',
            'lexbor',
            'selectolax',
            'selenium',
            'selenium-stealth',
            'pandas',
        ],
    install_requires=[
          'google-search-results>=2.4.2',
          'selectolax>=0.3.12',
          'parsel>=1.7.0',
          'selenium-stealth>=1.0.6',
          'pandas>=1.5.3',
          'webdriver-manager>=3.8.5' 
    ],
    project_urls={
        'Documentation': 'https://github.com/dimitryzub/scrape-google-scholar#example-usage-custom-backend',
        'Source': 'https://github.com/dimitryzub/scrape-google-scholar',
        'Tracker': 'https://github.com/dimitryzub/scrape-google-scholar/issues',
    },
)